@@ -1,3 +1,5 @@
+# type: ignore
+
 """
 .. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
 """
@@ -22,7 +24,7 @@ from pathvalidate import (
 )
 from pathvalidate._common import unprintable_ascii_chars
 from pathvalidate._filename import FileNameSanitizer, FileNameValidator
-from pathvalidate.handler import raise_error, return_null_string, return_timestamp
+from pathvalidate.handler import NullValueHandler, ReservedNameHandler, raise_error
 
 from ._common import (
     INVALID_FILENAME_CHARS,
@@ -34,6 +36,9 @@ from ._common import (
     VALID_PLATFORM_NAMES,
     WIN_RESERVED_FILE_NAMES,
     is_faker_installed,
+    platform_linux,
+    platform_macos,
+    platform_windows,
     randstr,
 )
 
@@ -57,11 +62,11 @@ class Test_FileSanitizer:
     )
     def test_normal_platform_auto(self, monkeypatch, test_platform, expected):
         if test_platform == "windows":
-            patch = lambda: "windows"
+            patch = platform_windows
         elif test_platform == "linux":
-            patch = lambda: "linux"
+            patch = platform_linux
         elif test_platform == "macos":
-            patch = lambda: "macos"
+            patch = platform_macos
         else:
             raise ValueError(f"unexpected test platform: {test_platform}")
 
@@ -69,6 +74,10 @@ class Test_FileSanitizer:
 
         assert FileNameSanitizer(255, platform="auto").platform == expected
 
+    def test_normal_additional_reserved_names(self):
+        sanitizer = FileNameSanitizer(additional_reserved_names=["abc"])
+        assert sanitizer.reserved_keywords == ("ABC",)
+
 
 class Test_FileNameValidator:
     @pytest.mark.parametrize(
@@ -77,11 +86,8 @@ class Test_FileNameValidator:
             [
                 "windows",
                 (
-                    "CON",
-                    "PRN",
                     "AUX",
                     "CLOCK$",
-                    "NUL",
                     "COM1",
                     "COM2",
                     "COM3",
@@ -91,6 +97,7 @@ class Test_FileNameValidator:
                     "COM7",
                     "COM8",
                     "COM9",
+                    "CON",
                     "LPT1",
                     "LPT2",
                     "LPT3",
@@ -100,6 +107,8 @@ class Test_FileNameValidator:
                     "LPT7",
                     "LPT8",
                     "LPT9",
+                    "NUL",
+                    "PRN",
                 ),
             ],
             ["linux", ()],
@@ -109,6 +118,16 @@ class Test_FileNameValidator:
     def test_normal_reserved_keywords(self, test_platform, expected):
         assert FileNameValidator(255, platform=test_platform).reserved_keywords == expected
 
+    def test_normal_additional_reserved_names(self):
+        sanitizer = FileNameValidator(additional_reserved_names=["abc", "efg.txt"])
+        assert "ABC" in sanitizer.reserved_keywords
+        assert "EFG.TXT" in sanitizer.reserved_keywords
+
+        sanitizer = FileNameValidator(platform="windows", additional_reserved_names=["CON"])
+        assert (
+            sanitizer.reserved_keywords == FileNameValidator(platform="windows").reserved_keywords
+        )
+
 
 class Test_validate_filename:
     VALID_CHARS = VALID_FILENAME_CHARS
@@ -160,7 +179,7 @@ class Test_validate_filename:
         value = "  "
         with pytest.raises(ValidationError) as e:
             validate_filename(value, platform=platform)
-            assert e.value.reason == ErrorReason.NULL
+            assert e.value.reason == ErrorReason.NULL_NAME
         assert not is_valid_filename(value, platform)
 
     @pytest.mark.parametrize(
@@ -188,10 +207,13 @@ class Test_validate_filename:
         if expected is None:
             validate_filename(value, min_len=min_len)
             assert is_valid_filename(value, min_len=min_len)
-        else:
-            with pytest.raises(ValidationError) as e:
-                validate_filename(value, min_len=min_len)
-            assert e.value.reason == expected
+            return
+
+        with pytest.raises(ValidationError) as e:
+            validate_filename(value, min_len=min_len)
+        assert e.value.reason == expected
+        assert e.value.fs_encoding
+        assert e.value.byte_count > 0
 
     @pytest.mark.parametrize(
         ["value", "platform", "max_len", "expected"],
@@ -212,6 +234,8 @@ class Test_validate_filename:
         with pytest.raises(ValidationError) as e:
             validate_filename(value, platform=platform, max_len=max_len)
         assert e.value.reason == expected
+        assert e.value.fs_encoding
+        assert e.value.byte_count > 0
 
     @pytest.mark.parametrize(
         ["value", "platform", "fs_encoding", "max_len", "expected"],
@@ -237,6 +261,8 @@ class Test_validate_filename:
         with pytest.raises(ValidationError) as e:
             validate_filename(value, **kwargs)
         assert e.value.reason == expected
+        assert e.value.fs_encoding
+        assert e.value.byte_count > 0
 
     @pytest.mark.parametrize(
         ["value", "min_len", "max_len", "expected"],
@@ -252,9 +278,10 @@ class Test_validate_filename:
         if expected is None:
             validate_filename(value, min_len=min_len, max_len=max_len)
             assert is_valid_filename(value, min_len=min_len, max_len=max_len)
-        else:
-            with pytest.raises(expected):
-                validate_filename(value, min_len=min_len, max_len=max_len)
+            return
+
+        with pytest.raises(expected):
+            validate_filename(value, min_len=min_len, max_len=max_len)
 
     @pytest.mark.skipif(not is_faker_installed(), reason="requires faker")
     @pytest.mark.parametrize(["locale"], [[None], ["ja_JP"]])
@@ -357,28 +384,45 @@ class Test_validate_filename:
             assert not is_valid_filename(value, platform=platform)
 
     @pytest.mark.parametrize(
-        ["platform", "value", "expected"],
+        ["value", "arn", "expected"],
         [
-            [win_abspath, platform, None]
+            ["abc", [], True],
+            ["abc", ["abc"], False],
+            ["Abc", ["abc"], False],
+            ["ABC", ["abc"], False],
+            ["abc.txt", ["abc.txt"], False],
+        ],
+    )
+    def test_normal_additional_reserved_names(self, value, arn, expected):
+        assert is_valid_filename(value, additional_reserved_names=arn) == expected
+
+    @pytest.mark.parametrize(
+        ["platform", "value", "expected", "reason"],
+        [
+            [win_abspath, platform, None, None]
             for win_abspath, platform in product(
                 ["linux", "macos", "posix"],
                 ["\\", "\\\\", "\\ ", "C:\\", "c:\\", "\\xyz", "\\xyz "],
             )
         ]
         + [
-            [win_abspath, platform, ValidationError]
+            [win_abspath, platform, ValidationError, ErrorReason.FOUND_ABS_PATH]
+            for win_abspath, platform in product(["windows", "universal"], ["\\\\", "C:\\", "c:\\"])
+        ]
+        + [
+            [win_abspath, platform, ValidationError, ErrorReason.INVALID_CHARACTER]
             for win_abspath, platform in product(
-                ["windows", "universal"], ["\\", "\\\\", "\\ ", "C:\\", "c:\\", "\\xyz", "\\xyz "]
+                ["windows", "universal"], ["\\", "\\ ", "\\xyz", "\\xyz "]
             )
         ],
     )
-    def test_win_abs_path(self, platform, value, expected):
+    def test_win_abs_path(self, platform, value, expected, reason):
         if expected is None:
             validate_filename(value, platform=platform)
         else:
             with pytest.raises(expected) as e:
                 validate_filename(value, platform=platform)
-            assert e.value.reason == ErrorReason.FOUND_ABS_PATH
+            assert e.value.reason == reason
 
     @pytest.mark.parametrize(
         ["value", "platform"],
@@ -410,7 +454,7 @@ class Test_validate_filename:
         assert e.value.reason == ErrorReason.INVALID_CHARACTER
         assert str(e.value) == (
             r"[PV1100] invalid characters found: invalids=('\r'), value='asdf\rsdf', "
-            "target-platform=Windows"
+            "platform=Windows"
         )  # noqa
 
     @pytest.mark.parametrize(
@@ -425,21 +469,6 @@ class Test_validate_filename:
             validate_filename(value)
         assert not is_valid_filename(value)
 
-    @pytest.mark.parametrize(
-        ["value", "expected"],
-        [
-            ["a" * 256, ValidationError],
-            [1, TypeError],
-            [True, TypeError],
-            [nan, TypeError],
-            [inf, TypeError],
-        ],
-    )
-    def test_exception(self, value, expected):
-        with pytest.raises(expected):
-            validate_filename(value)
-        assert not is_valid_filename(value)
-
 
 class Test_sanitize_filename:
     SANITIZE_CHARS = INVALID_WIN_FILENAME_CHARS + unprintable_ascii_chars
@@ -534,8 +563,10 @@ class Test_sanitize_filename:
         ],
     )
     def test_normal_null_value_handler(self, value):
-        assert sanitize_filename(value, null_value_handler=return_null_string) == ""
-        assert sanitize_filename(value, null_value_handler=return_timestamp) != ""
+        assert (
+            sanitize_filename(value, null_value_handler=NullValueHandler.return_null_string) == ""
+        )
+        assert sanitize_filename(value, null_value_handler=NullValueHandler.return_timestamp) != ""
         with pytest.raises(ValidationError):
             sanitize_filename(value, null_value_handler=raise_error)
 
@@ -592,6 +623,47 @@ class Test_sanitize_filename:
         assert filename == expected
         assert is_valid_filename(filename, platform=test_platform)
 
+    @pytest.mark.parametrize(
+        ["value", "reserved_name_handler", "expected"],
+        [
+            ["CON", ReservedNameHandler.add_trailing_underscore, "CON_"],
+            ["CON", ReservedNameHandler.add_leading_underscore, "_CON"],
+            ["CON", ReservedNameHandler.as_is, "CON"],
+        ],
+    )
+    def test_normal_reserved_name_handler(self, value, reserved_name_handler, expected):
+        for platform in ["windows", "universal"]:
+            assert (
+                sanitize_filename(
+                    value, platform=platform, reserved_name_handler=reserved_name_handler
+                )
+                == expected
+            )
+
+    def test_exception_reserved_name_handler(self):
+        for platform in ["windows", "universal"]:
+            with pytest.raises(ValidationError) as e:
+                sanitize_filename("CON", platform=platform, reserved_name_handler=raise_error)
+            assert e.value.reason == ErrorReason.RESERVED_NAME
+
+    @pytest.mark.parametrize(
+        ["value", "arn", "expected"],
+        [
+            ["abc", [], "abc"],
+            ["abc", ["abc"], "abc_"],
+        ],
+    )
+    def test_normal_additional_reserved_names(self, value, arn, expected):
+        for platform in ["windows", "universal"]:
+            assert (
+                sanitize_filename(
+                    value,
+                    platform=platform,
+                    additional_reserved_names=arn,
+                )
+                == expected
+            )
+
     @pytest.mark.parametrize(
         ["value", "check_reserved", "expected"],
         [
@@ -673,3 +745,20 @@ class Test_sanitize_filename:
         with pytest.raises(expected):
             sanitize_filename(value)
         assert not is_valid_filename(value)
+
+    @pytest.mark.parametrize(
+        ["value", "platform", "fs_encoding", "max_len", "expected"],
+        [
+            ["あ" * 85, "universal", "utf-8", 255, "あ" * 85],
+            ["あ" * 86, "universal", "utf-8", 255, "あ" * 85],
+            ["あ" * 126, "universal", "utf-16", 255, "あ" * 126],
+            ["あ" * 127, "universal", "utf-16", 255, "あ" * 126],
+        ],
+    )
+    def test_max_len_fs_encoding(self, value, platform, fs_encoding, max_len, expected):
+        kwargs = {
+            "platform": platform,
+            "max_len": max_len,
+            "fs_encoding": fs_encoding,
+        }
+        assert sanitize_filename(value, **kwargs) == expected
