@@ -1,3 +1,5 @@
+# type: ignore
+
 """
 .. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
 """
@@ -22,7 +24,7 @@ from pathvalidate import (
 )
 from pathvalidate._common import unprintable_ascii_chars
 from pathvalidate._filepath import FilePathSanitizer, FilePathValidator
-from pathvalidate.handler import raise_error, return_null_string, return_timestamp
+from pathvalidate.handler import NullValueHandler, ReservedNameHandler, raise_error
 
 from ._common import (
     INVALID_PATH_CHARS,
@@ -31,6 +33,9 @@ from ._common import (
     VALID_PATH_CHARS,
     WIN_RESERVED_FILE_NAMES,
     is_faker_installed,
+    platform_linux,
+    platform_macos,
+    platform_windows,
     randstr,
 )
 
@@ -52,11 +57,11 @@ class Test_FileSanitizer:
     )
     def test_normal_platform_auto(self, monkeypatch, test_platform, expected):
         if test_platform == "windows":
-            patch = lambda: "windows"
+            patch = platform_windows
         elif test_platform == "linux":
-            patch = lambda: "linux"
+            patch = platform_linux
         elif test_platform == "macos":
-            patch = lambda: "macos"
+            patch = platform_macos
         else:
             raise ValueError(f"unexpected test platform: {test_platform}")
 
@@ -64,6 +69,10 @@ class Test_FileSanitizer:
 
         assert FilePathSanitizer(255, platform="auto").platform == expected
 
+    def test_normal_additional_reserved_names(self):
+        sanitizer = FilePathSanitizer(additional_reserved_names=["abc"])
+        assert sanitizer.reserved_keywords == ("ABC",)
+
 
 class Test_FilePathValidator:
     @pytest.mark.parametrize(
@@ -78,6 +87,10 @@ class Test_FilePathValidator:
     def test_normal_reserved_keywords(self, test_platform, expected):
         assert FilePathValidator(255, platform=test_platform).reserved_keywords == expected
 
+    def test_normal_additional_reserved_names(self):
+        sanitizer = FilePathValidator(additional_reserved_names=["abc"])
+        assert "ABC" in sanitizer.reserved_keywords
+
 
 class Test_validate_filepath:
     VALID_CHARS = VALID_PATH_CHARS
@@ -93,6 +106,7 @@ class Test_validate_filepath:
         "C:\\Users",
         "C:\\",
         "\\Users",
+        r"\\fileserver\public",
     ]
 
     @pytest.mark.parametrize(
@@ -176,6 +190,9 @@ class Test_validate_filepath:
         with pytest.raises(ValidationError) as e:
             validate_filepath(value, min_len=min_len)
         assert e.value.reason == expected
+        assert e.value.fs_encoding
+        assert e.value.byte_count
+        assert e.value.byte_count > 0
 
     @pytest.mark.parametrize(
         ["value", "platform", "max_len", "expected"],
@@ -211,6 +228,9 @@ class Test_validate_filepath:
         with pytest.raises(ValidationError) as e:
             validate_filepath(value, **kwargs)
         assert e.value.reason == ErrorReason.INVALID_LENGTH
+        assert e.value.fs_encoding
+        assert e.value.byte_count
+        assert e.value.byte_count > 0
 
     @pytest.mark.parametrize(
         ["value", "platform", "fs_encoding", "max_len", "expected"],
@@ -236,6 +256,9 @@ class Test_validate_filepath:
         with pytest.raises(ValidationError) as e:
             validate_filepath(value, **kwargs)
         assert e.value.reason == expected
+        assert e.value.fs_encoding
+        assert e.value.byte_count
+        assert e.value.byte_count > 0
 
     @pytest.mark.parametrize(
         ["value", "min_len", "max_len", "expected"],
@@ -267,7 +290,7 @@ class Test_validate_filepath:
         [
             ["linux", "/a/b/c.txt", None],
             ["linux", "C:\\a\\b\\c.txt", ValidationError],
-            ["windows", "/a/b/c.txt", None],
+            ["windows", "/a/b/c.txt", ValidationError],
             ["windows", "C:\\a\\b\\c.txt", None],
             ["universal", "/a/b/c.txt", ValidationError],
             ["universal", "C:\\a\\b\\c.txt", ValidationError],
@@ -341,24 +364,40 @@ class Test_validate_filepath:
     @pytest.mark.parametrize(
         ["platform", "value"],
         [
-            ["windows", "period."],
-            ["windows", "space "],
-            ["windows", "space_and_period ."],
-            ["windows", "space_and_period. "],
             ["linux", "period."],
             ["linux", "space "],
             ["linux", "space_and_period. "],
-            ["universal", "period."],
-            ["universal", "space "],
-            ["universal", "space_and_period ."],
         ],
     )
     def test_normal_space_or_period_at_tail(self, platform, value):
         validate_filepath(value, platform=platform)
         assert is_valid_filepath(value, platform=platform)
 
+    @pytest.mark.parametrize(
+        ["platform", "value"],
+        [
+            ["windows", "period."],
+            ["windows", "space "],
+            ["windows", "space_and_period ."],
+            ["windows", "space_and_period. "],
+            ["universal", "period."],
+            ["universal", "space "],
+            ["universal", "space_and_period ."],
+        ],
+    )
+    def test_exception_space_or_period_at_tail(self, platform, value):
+        with pytest.raises(ValidationError):
+            validate_filepath(value, platform=platform)
+        assert not is_valid_filepath(value, platform=platform)
+
     @pytest.mark.skipif(not is_faker_installed(), reason="requires faker")
-    @pytest.mark.parametrize(["locale"], [[None], ["ja_JP"]])
+    @pytest.mark.parametrize(
+        ["locale"],
+        [
+            [None],
+            ["ja_JP"],
+        ],
+    )
     def test_locale_jp(self, locale):
         from faker import Factory
 
@@ -459,6 +498,16 @@ class Test_validate_filepath:
 
         assert not is_valid_filepath(value, platform=platform)
 
+    @pytest.mark.parametrize(
+        ["value", "arn", "expected"],
+        [
+            ["abc/efg.txt", ["abc"], False],
+            ["abc/efg.txt", ["efg.txt"], False],
+        ],
+    )
+    def test_normal_additional_reserved_names(self, value, arn, expected):
+        assert is_valid_filepath(value, additional_reserved_names=arn) == expected
+
     @pytest.mark.parametrize(
         ["value", "platform", "expected"],
         [
@@ -474,7 +523,7 @@ class Test_validate_filepath:
         assert e.value.reason == expected
         assert str(e.value) == (
             r"[PV1100] invalid characters found: invalids=('\r'), value='asdf\rsdf', "
-            "target-platform=Windows"
+            "platform=Windows"
         )  # noqa
 
     @pytest.mark.parametrize(
@@ -648,6 +697,28 @@ class Test_sanitize_filepath:
         assert filename == expected
         assert is_valid_filepath(filename, platform=test_platform)
 
+    @pytest.mark.parametrize(
+        ["value", "reserved_name_handler", "expected"],
+        [
+            ["CON", ReservedNameHandler.add_trailing_underscore, "CON_"],
+            ["CON", ReservedNameHandler.add_leading_underscore, "_CON"],
+            ["CON", ReservedNameHandler.as_is, "CON"],
+        ],
+    )
+    def test_normal_reserved_name_handler(self, value, reserved_name_handler, expected):
+        assert (
+            sanitize_filepath(
+                value, platform="windows", reserved_name_handler=reserved_name_handler
+            )
+            == expected
+        )
+
+    def test_exception_reserved_name_handler(self):
+        for platform in ["windows", "universal"]:
+            with pytest.raises(ValidationError) as e:
+                sanitize_filepath("CON", platform=platform, reserved_name_handler=raise_error)
+            assert e.value.reason == ErrorReason.RESERVED_NAME
+
     @pytest.mark.parametrize(
         ["value", "check_reserved", "expected"],
         [
@@ -660,6 +731,23 @@ class Test_sanitize_filepath:
             sanitize_filepath(value, platform="windows", check_reserved=check_reserved) == expected
         )
 
+    @pytest.mark.parametrize(
+        ["value", "arn", "expected"],
+        [
+            ["abc", ["abc"], "abc_"],
+        ],
+    )
+    def test_normal_additional_reserved_names(self, value, arn, expected):
+        for platform in ["windows", "universal"]:
+            assert (
+                sanitize_filepath(
+                    value,
+                    platform=platform,
+                    additional_reserved_names=arn,
+                )
+                == expected
+            )
+
     @pytest.mark.parametrize(
         ["value", "replace_text", "expected"],
         [
@@ -730,8 +818,10 @@ class Test_sanitize_filepath:
         ],
     )
     def test_normal_null_value_handler(self, value):
-        assert sanitize_filepath(value, null_value_handler=return_null_string) == ""
-        assert sanitize_filepath(value, null_value_handler=return_timestamp) != ""
+        assert (
+            sanitize_filepath(value, null_value_handler=NullValueHandler.return_null_string) == ""
+        )
+        assert sanitize_filepath(value, null_value_handler=NullValueHandler.return_timestamp) != ""
         with pytest.raises(ValidationError):
             sanitize_filepath(value, null_value_handler=raise_error)
 
@@ -825,7 +915,6 @@ class Test_sanitize_filepath:
             sanitize_filepath(value, platform=platform, validate_after_sanitize=True)
         assert e.value.reason == ErrorReason.INVALID_AFTER_SANITIZE
 
-    @pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6 or higher")
     @pytest.mark.parametrize(
         ["value", "expected"],
         [
